from ...utils import rest
from ...utils.api import from_api_json
from ...utils.checks import check_datetime, check_date
from ...utils.resource import Resource
from ..__transfer import _resource as _transfer_resource


class Log(Resource):
    """# transfer.Log object
    Every time a Transfer entity is modified, a corresponding transfer.Log
    is generated for the entity. This log is never generated by the
    user.
    ## Attributes:
    - id [string]: unique id returned when the log is created. ex: "5656565656565656"
    - transfer [Transfer]: Transfer entity to which the log refers to.
    - errors [list of strings]: list of errors linked to this BoletoPayment event.
    - type [string]: type of the Transfer event which triggered the log creation. ex: "processing" or "success"
    - created [datetime.datetime]: creation datetime for the log. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    """

    def __init__(self, id, created, type, errors, transfer):
        Resource.__init__(self, id=id)

        self.created = check_datetime(created)
        self.type = type
        self.errors = errors
        self.transfer = from_api_json(_transfer_resource, transfer)


_resource = {"class": Log, "name": "TransferLog"}


def get(id, user=None):
    """# Retrieve a specific transfer.Log
    Receive a single transfer.Log object previously created by the Stark Bank API by its id
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - transfer.Log object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def query(limit=None, after=None, before=None, types=None, transfer_ids=None, user=None):
    """# Retrieve transfer.Log's
    Receive a generator of transfer.Log objects previously created in the Stark Bank API
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - after [datetime.date or string, default None] date filter for objects created only after specified date. ex: datetime.date(2020, 3, 10)
    - before [datetime.date or string, default None] date filter for objects created only before specified date. ex: datetime.date(2020, 3, 10)
    - types [list of strings, default None]: filter retrieved objects by types. ex: "success" or "failed"
    - transfer_ids [list of strings, default None]: list of Transfer ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - list of transfer.Log objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        limit=limit,
        after=check_date(after),
        before=check_date(before),
        types=types,
        transfer_ids=transfer_ids,
        user=user,
    )





class Resource:

    def __init__(self, id):
        id = str(id) if id else None
        self.id = id

    def __repr__(self):
        return "{classname}[{id}]".format(
            classname=self.__class__.__name__,
            id=self.id,
        )

    def __str__(self):
        dict_ = {
            attribute: getattr(self, attribute)
            for attribute in dir(self)
            if not attribute.startswith('_') and not callable(getattr(self, attribute))
        }
        fields = u",\n\t".join(u"{key}={value}".format(key=key, value="\n\t".join(str(value).splitlines())) for key, value in dict_.items())
        string = u"{classname}(\n\t{fields}\n)".format(
            classname=self.__class__.__name__,
            fields=fields
        )
        return string



from ..utils import rest
from ..utils.resource import Resource


class BrcodePreview(Resource):

    """# BrcodePreview object
    A BrcodePreview is used to get information from a BR Code you received to check the informations before paying it.
    ## Attributes (return-only):
    - status [string]: Payment status. ex: "active", "paid", "canceled" or "unknown"
    - name [string]: Payment receiver name. ex: "Tony Stark"
    - tax_id [string]: Payment receiver tax ID. ex: "012.345.678-90"
    - bank_code [string]: Payment receiver bank code. ex: "20018183"
    - branch_code [string]: Payment receiver branch code. ex: "0001"
    - account_number [string]: Payment receiver account number. ex: "1234567"
    - account_type [string]: Payment receiver account type. ex: "checking"
    - allow_change [bool]: If True, the payment is able to receive amounts that are different from the nominal one. ex: True or False
    - amount [integer]: Value in cents that this payment is expecting to receive. If 0, any value is accepted. ex: 123 (= R$1,23)
    - reconciliation_id [string]: Reconciliation ID linked to this payment. ex: "txId", "payment-123"
    """

    def __init__(self, status, name, tax_id, bank_code, branch_code, account_number, account_type, allow_change, amount, reconciliation_id):
        self.status = status
        self.name = name
        self.tax_id = tax_id
        self.bank_code = bank_code
        self.branch_code = branch_code
        self.account_number = account_number
        self.account_type = account_type
        self.allow_change = allow_change
        self.amount = amount
        self.reconciliation_id = reconciliation_id


_resource = {"class": BrcodePreview, "name": "BrcodePreview"}


def query(brcodes, user=None):
    """# Retrieve BrcodePreviews
    Receive a generator of BrcodePreview objects previously created in the Stark Bank API
    ## Parameters (optional):
    - brcodes [list of strings]: List of brcodes to preview. ex: ["00020126580014br.gov.bcb.pix0136a629532e-7693-4846-852d-1bbff817b5a8520400005303986540510.005802BR5908T'Challa6009Sao Paulo62090505123456304B14A"]
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of BrcodePreview objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        brcodes=brcodes,
        limit=None,
        user=user,
    )



from ..utils import rest
from ..utils.resource import Resource


class Workspace(Resource):
    """# Workspace object
    Workspaces are bank accounts. They have independent balances, statements, operations and permissions.
    The only property that is shared between your workspaces is that they are linked to your organization,
    which carries your basic informations, such as tax ID, name, etc..
    ## Parameters (required):
    - username [string]: Simplified name to define the workspace URL. This name must be unique across all Stark Bank Workspaces. Ex: "starkbankworkspace"
    - name [string]: Full name that identifies the Workspace. This name will appear when people access the Workspace on our platform, for example. Ex: "Stark Bank Workspace"
    ## Attributes:
    - id [string, default None]: unique id returned when the workspace is created. ex: "5656565656565656"
    """

    def __init__(self, username, name, id=None):
        Resource.__init__(self, id=id)

        self.username = username
        self.name = name


_resource = {"class": Workspace, "name": "Workspace"}


def create(username, name, user=None):
    """# Create Workspace
    Send a Workspace for creation in the Stark Bank API
    ## Parameters (required):
    - username [string]: Simplified name to define the workspace URL. This name must be unique across all Stark Bank Workspaces. Ex: "starkbankworkspace"
    - name [string]: Full name that identifies the Workspace. This name will appear when people access the Workspace on our platform, for example. Ex: "Stark Bank Workspace"
    ## Parameters (optional):
    - user [Organization object, default None]: Organization object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Workspace object with updated attributes
    """
    return rest.post_single(resource=_resource, entity=Workspace(username=username, name=name), user=user)


def get(id, user=None):
    """# Retrieve a specific Workspace
    Receive a single Workspace object previously created in the Stark Bank API by passing its id
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Workspace object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def query(limit=None, username=None, ids=None, user=None):
    """# Retrieve Workspaces
    Receive a generator of Workspace objects previously created in the Stark Bank API.
    If no filters are passed and the user is an Organization, all of the Organization Workspaces
    will be retrieved.
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - username [string]: query by the simplified name that defines the workspace URL. This name is always unique across all Stark Bank Workspaces. Ex: "starkbankworkspace"
    - ids [list of strings, default None]: list of ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of Workspace objects with updated attributes
    """
    return rest.get_list(resource=_resource, limit=limit, username=username, ids=ids, user=user)



from ..utils import rest
from ..utils.api import from_api_json
from ..utils.resource import Resource
from ..utils.checks import check_datetime, check_date
from ..transfer.__transfer import Transfer
from ..brcodepayment.__brcodepayment import BrcodePayment
from ..boletopayment.__boletopayment import BoletoPayment
from ..utilitypayment.__utilitypayment import UtilityPayment
from ..brcodepayment.__brcodepayment import BrcodePayment
from ..transaction.__transaction import Transaction
from ..transfer.__transfer import _resource as _transfer_resource
from ..brcodepayment.__brcodepayment import _resource as _brcode_payment_resource
from ..boletopayment.__boletopayment import _resource as _boleto_payment_resource
from ..utilitypayment.__utilitypayment import _resource as _utility_payment_resource
from ..brcodepayment.__brcodepayment import _resource as _brcode_payment_resource
from ..transaction.__transaction import _resource as _transaction_resource


class PaymentRequest(Resource):
    """# PaymentRequest object
    A PaymentRequest is an indirect request to access a specific cash-out service
    (such as Transfers, BoletoPayments, etc.) which goes through the cost center
    approval flow on our website. To emit a PaymentRequest, you must direct it to
    a specific cost center by its ID, which can be retrieved on our website at the
    cost center page.
    ## Parameters (required):
    - center_id [string]: target cost center ID. ex: "5656565656565656"
    - payment [Transfer, BoletoPayment, UtilityPayment, BrcodePayment, Transaction or dictionary]: payment entity that should be approved and executed.
    ## Parameters (conditionally required):
    - type [string]: payment type, inferred from the payment parameter if it is not a dictionary. ex: "transfer", "boleto-payment"
    ## Parameters (optional):
    - due [datetime.date or string, default today]: Payment target date in ISO format. ex: 2020-04-30
    - tags [list of strings]: list of strings for tagging
    ## Attributes (return-only):
    - id [string, default None]: unique id returned when a PaymentRequest is created. ex: "5656565656565656"
    - amount [integer, default None]: PaymentRequest amount. ex: 100000 = R$1.000,00
    - status [string, default None]: current PaymentRequest status. ex: "pending" or "approved"
    - actions [list of dictionaries, default None]: list of actions that are affecting this PaymentRequest. ex: [{"type": "member", "id": "56565656565656, "action": "requested"}]
    - updated [datetime.datetime, default None]: latest update datetime for the PaymentRequest. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    - created [datetime.datetime, default None]: creation datetime for the PaymentRequest. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    """

    def __init__(self, center_id, payment, type=None, due=None, tags=None, id=None, amount=None,
                 status=None, actions=None, updated=None, created=None):
        Resource.__init__(self, id=id)

        self.center_id = center_id
        self.due = due
        self.tags = tags
        self.amount = amount
        self.status = status
        self.actions = actions
        self.updated = updated
        self.created = check_datetime(created)

        self.payment, self.type = _parse_payment(payment=payment, type=type)


def _parse_payment(payment, type):
    if isinstance(payment, dict):
        try:
            return from_api_json(*({
                "transfer": _transfer_resource,
                "transaction": _transaction_resource,
                "boleto-payment": _boleto_payment_resource,
                "brcode-payment": _brcode_payment_resource,
                "utility-payment": _utility_payment_resource,
                "brcode-payment": _brcode_payment_resource,
            }[type], payment)), type
        except KeyError:
            return payment, type

    if type:
        return payment, type

    if isinstance(payment, Transfer):
        return payment, "transfer"
    if isinstance(payment, Transaction):
        return payment, "transaction"
    if isinstance(payment, BrcodePayment):
        return payment, "brcode-payment"
    if isinstance(payment, BoletoPayment):
        return payment, "boleto-payment"
    if isinstance(payment, UtilityPayment):
        return payment, "utility-payment"
    if isinstance(payment, BrcodePayment):
        return payment, "brcode-payment"

    raise Exception(
        "payment must be either "
        "a dictionary"
        ", a starkbank.Transfer"
        ", a starkbank.Transaction"
        ", a starkbank.BrcodePayment"
        ", a starkbank.BoletoPayment"
        ", a starkbank.UtilityPayment"
        " or a starkbank.BrcodePayment"
        ", but not a {}".format(type(payment))
    )


_resource = {"class": PaymentRequest, "name": "PaymentRequest"}


def create(requests, user=None):
    """# Create PaymentRequests
    Send a list of PaymentRequest objects for creation in the Stark Bank API
    ## Parameters (required):
    - requests [list of PaymentRequest objects]: list of PaymentRequest objects to be created in the API
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - list of PaymentRequest objects with updated attributes
    """
    return rest.post_multi(resource=_resource, entities=requests, user=user)


def query(center_id, limit=None, after=None, before=None, sort=None, status=None, type=None, tags=None, ids=None, user=None):
    """# Retrieve PaymentRequests
    Receive a generator of PaymentRequest objects previously created by this user in the Stark Bank API
    ## Parameters (required):
    - center_id [string]: target cost center ID. ex: "5656565656565656"
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - after [datetime.date or string, default None]: date filter for objects created or updated only after specified date. ex: datetime.date(2020, 3, 10)
    - before [datetime.date or string, default None]: date filter for objects created or updated only before specified date. ex: datetime.date(2020, 3, 10)
    - sort [string, default "-created"]: sort order considered in response. Valid options are "-created" or "-due".
    - status [string, default None]: filter for status of retrieved objects. ex: "success" or "failed"
    - type [string, default None]: payment type, inferred from the payment parameter if it is not a dictionary. ex: "transfer", "boleto-payment"
    - tags [list of strings, default None]: tags to filter retrieved objects. ex: ["tony", "stark"]
    - ids [list of strings, default None]: list of ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of PaymentRequest objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        center_id=center_id,
        limit=limit,
        after=check_date(after),
        before=check_date(before),
        sort=sort,
        status=status,
        type=type,
        tags=tags,
        ids=ids,
        user=user,
    )



from ellipticcurve.privateKey import PrivateKey
from os import makedirs, path as os_path


def create(path=None):
    """# Generate a new key pair
    Generates a secp256k1 ECDSA private/public key pair to be used in the API authentications
    ## Parameters (optional):
        path [string]: path to save the keys .pem files. No files will be saved if this parameter isn't provided
    ## Return:
        private and public key pems
    """

    private = PrivateKey()
    public = private.publicKey()

    private_pem = private.toPem()
    public_pem = public.toPem()

    if path is not None:
        if not os_path.exists(path):
            makedirs(path)
        with open(os_path.join(path, "private-key.pem"), "w") as file:
            file.write(private_pem)
        with open(os_path.join(path, "public-key.pem"), "w") as file:
            file.write(public_pem)

    return private_pem, public_pem



from ..utils import rest
from ..utils.resource import Resource


class Webhook(Resource):
    """# Webhook subscription object
    A Webhook is used to subscribe to notification events on a user-selected endpoint.
    Currently available services for subscription are transfer, boleto, boleto-holmes,
    boleto-payment, brcode-payment, utility-payment, deposit and invoice.
    ## Parameters (required):
    - url [string]: Url that will be notified when an event occurs.
    - subscriptions [list of strings]: list of any non-empty combination of the available services. ex: ["transfer", "invoice", "deposit"]
    ## Attributes:
    - id [string, default None]: unique id returned when the webhook is created. ex: "5656565656565656"
    """

    def __init__(self, url, subscriptions, id=None):
        Resource.__init__(self, id=id)

        self.url = url
        self.subscriptions = subscriptions


_resource = {"class": Webhook, "name": "Webhook"}


def create(url, subscriptions, user=None):
    """# Create Webhook subscription
    Send a single Webhook subscription for creation in the Stark Bank API
    ## Parameters (required):
    - url [string]: url to which notification events will be sent to. ex: "https://webhook.site/60e9c18e-4b5c-4369-bda1-ab5fcd8e1b29"
    - subscriptions [list of strings]: list of any non-empty combination of the available services. ex: ["transfer", "boleto-payment"]
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Webhook object with updated attributes
    """
    return rest.post_single(resource=_resource, entity=Webhook(url=url, subscriptions=subscriptions), user=user)


def get(id, user=None):
    """# Retrieve a specific Webhook subscription
    Receive a single Webhook subscription object previously created in the Stark Bank API by its id
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Webhook object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def query(limit=None, user=None):
    """# Retrieve Webhook subcriptions
    Receive a generator of Webhook subcription objects previously created in the Stark Bank API
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of Webhook objects with updated attributes
    """
    return rest.get_list(resource=_resource, limit=limit, user=user)


def delete(id, user=None):
    """# Delete a Webhook subscription entity
    Delete a Webhook subscription entity previously created in the Stark Bank API
    ## Parameters (required):
    - id [string]: Webhook unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - deleted Webhook object
    """
    return rest.delete_id(resource=_resource, id=id, user=user)



from ..utils import rest
from ..utils.checks import check_datetime, check_date, check_datetime_or_date
from ..utils.resource import Resource


class Transfer(Resource):
    """# Transfer object
    When you initialize a Transfer, the entity will not be automatically
    created in the Stark Bank API. The 'create' function sends the objects
    to the Stark Bank API and returns the list of created objects.
    ## Parameters (required):
    - amount [integer]: amount in cents to be transferred. ex: 1234 (= R$ 12.34)
    - name [string]: receiver full name. ex: "Anthony Edward Stark"
    - tax_id [string]: receiver tax ID (CPF or CNPJ) with or without formatting. ex: "01234567890" or "20.018.183/0001-80"
    - bank_code [string]: code of the receiver bank institution in Brazil. If an ISPB (8 digits) is informed, a PIX transfer will be created, else a TED will be issued. ex: "20018183" or "341"
    - branch_code [string]: receiver bank account branch. Use '-' in case there is a verifier digit. ex: "1357-9"
    - account_number [string]: Receiver bank account number. Use '-' before the verifier digit. ex: "876543-2"
    ## Parameters (optional):
    - account_type [string, default "checking"]: Receiver bank account type. This parameter only has effect on Pix Transfers. ex: "checking", "savings" or "salary"
    - external_id [string, default None]: url safe string that must be unique among all your transfers. Duplicated external_ids will cause failures. By default, this parameter will block any transfer that repeats amount and receiver information on the same date. ex: "my-internal-id-123456"
    - scheduled [datetime.date, datetime.datetime or string, default now]: date or datetime when the transfer will be processed. May be pushed to next business day if necessary. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    - tags [list of strings]: list of strings for reference when searching for transfers. ex: ["employees", "monthly"]
    ## Attributes (return-only):
    - id [string, default None]: unique id returned when the transfer is created. ex: "5656565656565656"
    - fee [integer, default None]: fee charged when transfer is created. ex: 200 (= R$ 2.00)
    - status [string, default None]: current transfer status. ex: "success" or "failed"
    - transaction_ids [list of strings, default None]: ledger transaction ids linked to this transfer (if there are two, second is the chargeback). ex: ["19827356981273"]
    - created [datetime.datetime, default None]: creation datetime for the transfer. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    - updated [datetime.datetime, default None]: latest update datetime for the transfer. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    """

    def __init__(self, amount, name, tax_id, bank_code, branch_code, account_number, account_type=None, external_id=None, scheduled=None, transaction_ids=None, fee=None, tags=None, status=None, id=None, created=None, updated=None):
        Resource.__init__(self, id=id)

        self.tax_id = tax_id
        self.amount = amount
        self.name = name
        self.bank_code = bank_code
        self.branch_code = branch_code
        self.account_number = account_number
        self.account_type = account_type
        self.external_id = external_id
        self.scheduled = check_datetime_or_date(scheduled)
        self.tags = tags
        self.fee = fee
        self.status = status
        self.created = check_datetime(created)
        self.updated = check_datetime(updated)
        self.transaction_ids = transaction_ids


_resource = {"class": Transfer, "name": "Transfer"}


def create(transfers, user=None):
    """# Create Transfers
    Send a list of Transfer objects for creation in the Stark Bank API
    ## Parameters (required):
    - transfers [list of Transfer objects]: list of Transfer objects to be created in the API
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - list of Transfer objects with updated attributes
    """
    return rest.post_multi(resource=_resource, entities=transfers, user=user)


def get(id, user=None):
    """# Retrieve a specific Transfer
    Receive a single Transfer object previously created in the Stark Bank API by its id
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Transfer object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def delete(id, user=None):
    """# Delete a Transfer entity
    Delete a Transfer entity previously created in the Stark Bank API
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - deleted Transfer object
    """
    return rest.delete_id(resource=_resource, id=id, user=user)


def pdf(id, user=None):
    """# Retrieve a specific Transfer pdf file
    Receive a single Transfer pdf receipt file generated in the Stark Bank API by its id.
    Only valid for transfers with "processing" and "success" status.
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Transfer pdf file
    """
    return rest.get_pdf(resource=_resource, id=id, user=user)


def query(limit=None, after=None, before=None, transaction_ids=None, status=None, tax_id=None, sort=None, tags=None, ids=None, user=None):
    """# Retrieve Transfers
    Receive a generator of Transfer objects previously created in the Stark Bank API
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - after [datetime.date or string, default None]: date filter for objects created or updated only after specified date. ex: datetime.date(2020, 3, 10)
    - before [datetime.date or string, default None]: date filter for objects created or updated only before specified date. ex: datetime.date(2020, 3, 10)
    - transaction_ids [list of strings, default None]: list of transaction IDs linked to the desired transfers. ex: ["5656565656565656", "4545454545454545"]
    - status [string, default None]: filter for status of retrieved objects. ex: "success" or "failed"
    - tax_id [string, default None]: filter for transfers sent to the specified tax ID. ex: "012.345.678-90"
    - sort [string, default "-created"]: sort order considered in response. Valid options are "created", "-created", "updated" or "-updated".
    - tags [list of strings, default None]: tags to filter retrieved objects. ex: ["tony", "stark"]
    - ids [list of strings, default None]: list of ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of Transfer objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        limit=limit,
        after=check_date(after),
        before=check_date(before),
        transaction_ids=transaction_ids,
        status=status,
        tax_id=tax_id,
        sort=sort,
        tags=tags,
        ids=ids,
        user=user,
    )



from datetime import datetime, date, timedelta


def check_environment(environment):
    from ..environment import Environment
    environments = Environment.values()
    assert environment in environments, "Select a valid environment: {}".format(", ".join(environments))
    return environment


def check_private_key(pem):
    from ellipticcurve.privateKey import PrivateKey
    try:
        assert PrivateKey.fromPem(pem).curve.name == "secp256k1"
    except:
        raise Exception("Private-key must be valid secp256k1 ECDSA string in pem format")
    return pem


def check_user(user):
    from ..user.__user import User
    assert isinstance(user, User), "A user is required to access our API. Check our README: https://github.com/starkbank/sdk-python/"
    return user


def check_language(language):
    accepted_languages = ["en-US", "pt-BR"]
    assert language in accepted_languages, "Language must be one from {}".format(accepted_languages)
    return language


def check_datetime_or_date(data):
    if data is None:
        return None

    if type(data) == datetime:
        return data

    if isinstance(data, date):
        return data

    data, dt_type = check_datetime_string(data)

    return data.date() if dt_type == date else data


def check_datetime(data):
    if data is None:
        return None

    if type(data) == datetime:
        return data

    if isinstance(data, date):
        return datetime(data.year, data.month, data.day)

    return check_datetime_string(data)[0]


def check_date(data):
    if data is None:
        return None

    if isinstance(data, datetime):
        return data.date()

    if isinstance(data, date):
        return data

    data, type = check_datetime_string(data)

    return data.date() if type == date else data


def check_timedelta(data):
    if data is None:
        return None

    if isinstance(data, timedelta):
        return data

    try:
        return timedelta(seconds=data)
    except:
        raise TypeError(
            "invalid timedelta {data}, please use an integer in seconds or a datetime.timedelta object".format(data=data)
        )


def check_datetime_string(data):
    data = str(data)

    try:
        return datetime.strptime(data, "%Y-%m-%d"), date
    except:
        pass

    try:
        return datetime.strptime(data, "%Y-%m-%dT%H:%M:%S.%f+00:00"), datetime
    except:
        pass

    try:
        return datetime.strptime(data, "%Y-%m-%dT%H:%M:%S+00:00"), datetime
    except:
        pass

    raise RuntimeError("invalid datetime string " + data)



from datetime import date, datetime, timedelta
from .resource import Resource
from .case import camel_to_kebab, snake_to_camel, camel_to_snake


def api_json(entity):
    if isinstance(entity, dict):
        return cast_json_to_api_format(entity)
    json = {
        attribute: getattr(entity, attribute)
        for attribute in dir(entity)
        if not attribute.startswith('_') and not callable(getattr(entity, attribute))
    }
    return cast_json_to_api_format(json)


def cast_json_to_api_format(json):
    return {snake_to_camel(k): cast_values(v) for k, v in json.items() if v is not None}


def cast_values(value):
    if type(value) == datetime:
        return value.strftime("%Y-%m-%dT%H:%M:%S+00:00")

    if isinstance(value, date):
        return value.strftime("%Y-%m-%d")

    if isinstance(value, timedelta):
        return int(value.total_seconds())

    if isinstance(value, Resource):
        return api_json(value)

    if not isinstance(value, list):
        return value

    casted_values = []
    for v in value:
        if isinstance(v, dict):
            casted_values.append(cast_json_to_api_format(v))
            continue
        casted_values.append(v)
    return casted_values


def from_api_json(resource, json):
    snakes = {camel_to_snake(k): v for k, v in json.items()}

    params = set(resource["class"].__init__.__code__.co_varnames) - {"self"}

    snakes = {k: v for k, v in snakes.items() if k in params}
    for param in params - set(snakes):
        snakes[param] = None

    return resource["class"](**snakes)


def endpoint(resource):
    return camel_to_kebab(resource["name"]).replace("-log", "/log")


def last_name(resource):
    return camel_to_kebab(resource["name"]).split("-")[-1]


def last_name_plural(resource):
    base = last_name(resource)
    if base.endswith("s"):
        return base
    if base.endswith("y") and not base.endswith("ey"):
        return "{name}ies".format(name=base[:-1])
    return "{name}s".format(name=base)



from requests import get, post, delete, patch
from ..utils.api import endpoint, last_name, last_name_plural, api_json, from_api_json, cast_json_to_api_format
from ..utils.request import fetch


def get_list(resource, limit=None, user=None, **kwargs):
    query = {"limit": min(limit, 100) if limit else limit}
    query.update(kwargs)

    while True:
        json = fetch(method=get, path=endpoint(resource), query=query, user=user).json()
        entities = json[last_name_plural(resource)]

        for entity in entities:
            yield from_api_json(resource, entity)

        if limit:
            limit -= 100
            query["limit"] = min(limit, 100)

        cursor = json.get("cursor")
        query["cursor"] = cursor
        if not cursor or (limit is not None and limit <= 0):
            break


def get_id(resource, id, user=None):
    json = fetch(method=get, path="{endpoint}/{id}".format(endpoint=endpoint(resource), id=id), user=user).json()
    entity = json[last_name(resource)]
    return from_api_json(resource, entity)


def get_pdf(resource, id, user=None, **kwargs):
    return fetch(method=get, path="{endpoint}/{id}/pdf".format(endpoint=endpoint(resource), id=id), query=kwargs, user=user).content


def get_qrcode(resource, id, user=None, **kwargs):
    return fetch(method=get, path="{endpoint}/{id}/qrcode".format(endpoint=endpoint(resource), id=id), query=kwargs, user=user).content


def post_multi(resource, entities, user=None):
    json = fetch(method=post, path=endpoint(resource), user=user, payload={
        last_name_plural(resource): [api_json(entity) for entity in entities]
    }).json()
    entities = json[last_name_plural(resource)]
    return [from_api_json(resource, entity) for entity in entities]


def post_single(resource, entity, user=None):
    payload = api_json(entity)
    json = fetch(method=post, path=endpoint(resource), user=user, payload=payload).json()
    entity_json = json[last_name(resource)]
    return from_api_json(resource, entity_json)


def delete_id(resource, id, user=None):
    json = fetch(method=delete, path="{endpoint}/{id}".format(endpoint=endpoint(resource), id=id), user=user).json()
    entity = json[last_name(resource)]
    return from_api_json(resource, entity)


def patch_id(resource, id, user=None, **payload):
    payload = cast_json_to_api_format(payload)
    json = fetch(method=patch, path="{endpoint}/{id}".format(endpoint=endpoint(resource), id=id), payload=payload, user=user).json()
    entity = json[last_name(resource)]
    return from_api_json(resource, entity)



from ellipticcurve.privateKey import PrivateKey
from ..utils.checks import check_private_key, check_environment
from ..utils.resource import Resource


class User(Resource):

    def __init__(self, id=None, private_key=None, environment=None):
        Resource.__init__(self, id=id)
        self.pem = check_private_key(private_key)
        self.environment = check_environment(environment)

    def private_key(self):
        return PrivateKey.fromPem(self.pem)



from ..utils import rest
from ..utils.checks import check_datetime, check_date
from ..utils.resource import Resource


class BoletoHolmes(Resource):
    """# BoletoHolmes object
    When you initialize a BoletoHolmes, the entity will not be automatically
    created in the Stark Bank API. The 'create' function sends the objects
    to the Stark Bank API and returns the list of created objects.
    ## Parameters (required):
    - boleto_id [string]: investigated boleto entity ID. ex: "5656565656565656"
    ## Parameters (optional):
    - tags [list of strings]: list of strings for tagging
    ## Attributes (return-only):
    - id [string, default None]: unique id returned when holmes is created. ex: "5656565656565656"
    - status [string, default None]: current holmes status. ex: "solving" or "solved"
    - result [string, default None]: result of boleto status investigation. ex: "paid" or "cancelled"
    - created [datetime.datetime, default None]: creation datetime for the holmes. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    - updated [datetime.datetime, default None]: latest update datetime for the holmes. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    """

    def __init__(self, boleto_id, tags=None, status=None, result=None, created=None, updated=None, id=None):
        Resource.__init__(self, id=id)

        self.boleto_id = boleto_id
        self.tags = tags
        self.status = status
        self.result = result
        self.created = check_datetime(created)
        self.updated = check_datetime(updated)


_resource = {"class": BoletoHolmes, "name": "BoletoHolmes"}


def create(holmes, user=None):
    """# Create BoletoHolmes
    Send a list of BoletoHolmes objects for creation in the Stark Bank API
    ## Parameters (required):
    - holmes [list of BoletoHolmes objects]: list of BoletoHolmes objects to be created in the API
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - list of BoletoHolmes objects with updated attributes
    """
    return rest.post_multi(resource=_resource, entities=holmes, user=user)


def get(id, user=None):
    """# Retrieve a specific BoletoHolmes
    Receive a single BoletoHolmes object previously created by the Stark Bank API by passing its id
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - BoletoHolmes object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def query(limit=None, after=None, before=None, tags=None, ids=None, status=None, boleto_id=None, user=None):
    """# Retrieve BoletoHolmes
    Receive a generator of BoletoHolmes objects previously created in the Stark Bank API
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - after [datetime.date or string, default None] date filter for objects created only after specified date. ex: datetime.date(2020, 3, 10)
    - before [datetime.date or string, default None] date filter for objects created only before specified date. ex: datetime.date(2020, 3, 10)
    - tags [list of strings, default None]: tags to filter retrieved objects. ex: ["tony", "stark"]
    - ids [list of strings, default None]: list of ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - status [string, default None]: filter for status of retrieved objects. ex: "solved"
    - boleto_id [string, default None]: filter for holmes that investigate a specific boleto by its ID. ex: "5656565656565656"
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of BoletoHolmes objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        limit=limit,
        after=check_date(after),
        before=check_date(before),
        tags=tags,
        ids=ids,
        status=status,
        boleto_id=boleto_id,
        user=user,
    )



from sys import version_info as python_version
from ellipticcurve.ecdsa import Ecdsa
from json import dumps, loads
from time import time
from ..error import InternalServerError, InputErrors, UnknownError
from ..environment import Environment
from .checks import check_user, check_language
from .url import urlencode
import starkbank


class Response:

    def __init__(self, status, content):
        self.status = status
        self.content = content

    def json(self):
        return loads(self.content.decode("utf-8"))


def fetch(method, path, payload=None, query=None, user=None, version="v2"):
    user = check_user(user or starkbank.user)
    language = check_language(starkbank.language)

    url = {
        Environment.production:  "https://api.starkbank.com/",
        Environment.sandbox:     "https://sandbox.api.starkbank.com/",
    }[user.environment] + version

    url = "{base_url}/{path}{query}".format(base_url=url, path=path, query=urlencode(query))

    agent = "Python-{major}.{minor}.{micro}-SDK-{sdk_version}".format(
        major=python_version.major,
        minor=python_version.minor,
        micro=python_version.micro,
        sdk_version=starkbank.version,
    )

    access_time = str(time())
    body = dumps(payload) if payload else ""
    message = "{access_id}:{access_time}:{body}".format(access_id=user.access_id(), access_time=access_time, body=body)
    signature = Ecdsa.sign(message=message, privateKey=user.private_key()).toBase64()

    try:
        request = method(
            url=url,
            data=body,
            headers={
                "Access-Id": user.access_id(),
                "Access-Time": access_time,
                "Access-Signature": signature,
                "Content-Type": "application/json",
                "User-Agent": agent,
                "Accept-Language": language,
            },
            timeout=starkbank.timeout
        )
    except Exception as exception:
        raise UnknownError("{}: {}".format(exception.__class__.__name__, str(exception)))

    response = Response(status=request.status_code, content=request.content)

    if response.status == 500:
        raise InternalServerError()
    if response.status == 400:
        raise InputErrors(response.json()["errors"])
    if response.status != 200:
        raise UnknownError(response.content)

    return response



from .__user import User


class Project(User):
    """# Project object
    The Project object is an authentication entity for the SDK that is permanently
    linked to a specific Workspace.
    All requests to the Stark Bank API must be authenticated via an SDK user,
    which must have been previously created at the Stark Bank website
    [https://sandbox.web.starkbank.com] or [https://web.starkbank.com]
    before you can use it in this SDK. Projects may be passed as the user parameter on
    each request or may be defined as the default user at the start (See README).
    ## Parameters (required):
    - id [string]: unique id required to identify project. ex: "5656565656565656"
    - private_key [EllipticCurve.Project()]: PEM string of the private key linked to the project. ex: "-----BEGIN PUBLIC KEY-----\nMFYwEAYHKoZIzj0CAQYFK4EEAAoDQgAEyTIHK6jYuik6ktM9FIF3yCEYzpLjO5X/\ntqDioGM+R2RyW0QEo+1DG8BrUf4UXHSvCjtQ0yLppygz23z0yPZYfw==\n-----END PUBLIC KEY-----"
    - environment [string]: environment where the project is being used. ex: "sandbox" or "production"
    ## Attributes (return-only):
    - name [string, default ""]: project name. ex: "MyProject"
    - allowed_ips [list of strings]: list containing the strings of the ips allowed to make requests on behalf of this project. ex: ["190.190.0.50"]
    - pem [string]: private key in pem format. ex: "-----BEGIN PUBLIC KEY-----\nMFYwEAYHKoZIzj0CAQYFK4EEAAoDQgAEyTIHK6jYuik6ktM9FIF3yCEYzpLjO5X/\ntqDioGM+R2RyW0QEo+1DG8BrUf4UXHSvCjtQ0yLppygz23z0yPZYfw==\n-----END PUBLIC KEY-----"
    """

    def __init__(self, id, environment, private_key, name="", allowed_ips=None):
        self.name = name
        self.allowed_ips = allowed_ips

        User.__init__(
            self,
            id=id,
            private_key=private_key,
            environment=environment,
        )

    def access_id(self):
        return "project/{id}".format(id=self.id)





class Error(Exception):

    def __init__(self, code, message):
        super(Exception, self).__init__("{code}: {message}".format(code=code, message=message))
        self.code = code
        self.message = message


class InputErrors(Exception):

    def __init__(self, content):
        self.errors = [Error(code=error["code"], message=error["message"]) for error in content]
        super(Exception, self).__init__(str(content))


class InternalServerError(Exception):

    def __init__(self, message="Houston, we have a problem."):
        super(Exception, self).__init__(message)


class UnknownError(Exception):

    def __init__(self, message):
        super(Exception, self).__init__("Unknown exception encountered: {}".format(message))


class InvalidSignatureError(Exception):
    pass



from .api import cast_json_to_api_format


def urlencode(params):
    params = cast_json_to_api_format(params) if params else {}
    params = {k: ",".join([str(value) for value in v]) if isinstance(v, (tuple, list, set)) else v for k, v in params.items()}
    return "?" + "&".join("{k}={v}".format(k=k, v=v) for k, v in params.items()) if params else ""



from ..utils import rest
from ..utils.checks import check_datetime, check_date
from ..utils.resource import Resource


class BrcodePayment(Resource):

    """# BrcodePayment object
    When you initialize a BrcodePayment, the entity will not be automatically
    created in the Stark Bank API. The 'create' function sends the objects
    to the Stark Bank API and returns the list of created objects.
    ## Parameters (required):
    - brcode [string]: String loaded directly from the QR Code or copied from the invoice. ex: "00020126580014br.gov.bcb.pix0136a629532e-7693-4846-852d-1bbff817b5a8520400005303986540510.005802BR5908T'Challa6009Sao Paulo62090505123456304B14A"
    - tax_id [string]: receiver tax ID (CPF or CNPJ) with or without formatting. ex: "01234567890" or "20.018.183/0001-80"
    - description [string]: Text to be displayed in your statement (min. 10 characters). ex: "payment ABC"
    ## Parameters (conditionally required):
    - amount [int, default None]: If the BRCode does not provide an amount, this parameter is mandatory, else it is optional. ex: 23456 (= R$ 234.56)
    ## Parameters (optional):
    - scheduled [datetime.date, datetime.datetime or string, default now]: payment scheduled date or datetime. ex: datetime.datetime(2020, 3, 10, 15, 17, 3)
    - tags [list of strings, default None]: list of strings for tagging
    ## Attributes (return-only):
    - id [string, default None]: unique id returned when payment is created. ex: "5656565656565656"
    - name [string]: receiver name. ex: "Jon Snow"
    - status [string, default None]: current payment status. ex: "success" or "failed"
    - type [string, default None]: brcode type. ex: "static" or "dynamic"
    - fee [integer, default None]: fee charged by this brcode payment. ex: 50 (= R$ 0.50)
    - updated [datetime.datetime, default None]: latest update datetime for the payment. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    - created [datetime.datetime, default None]: creation datetime for the payment. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    """

    def __init__(self, brcode, tax_id, description, amount=None, scheduled=None, tags=None, id=None, name=None,
                 status=None, type=None, fee=None, updated=None, created=None):
        Resource.__init__(self, id=id)

        self.brcode = brcode
        self.tax_id = tax_id
        self.description = description
        self.tags = tags
        self.scheduled = check_date(scheduled)
        self.name = name
        self.status = status
        self.type = type
        self.amount = amount
        self.fee = fee
        self.updated = check_datetime(updated)
        self.created = check_datetime(created)


_resource = {"class": BrcodePayment, "name": "BrcodePayment"}


def create(payments, user=None):
    """# Create BrcodePayments
    Send a list of BrcodePayment objects for creation in the Stark Bank API
    ## Parameters (required):
    - payments [list of BrcodePayment objects]: list of BrcodePayment objects to be created in the API
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - list of BrcodePayment objects with updated attributes
    """
    return rest.post_multi(resource=_resource, entities=payments, user=user)


def get(id, user=None):
    """# Retrieve a specific BrcodePayment
    Receive a single BrcodePayment object previously created by the Stark Bank API by its id
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - BrcodePayment object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def pdf(id, user=None):
    """# Retrieve a specific BrcodePayment pdf file
    Receive a single BrcodePayment pdf receipt file generated in the Stark Bank API by its id.
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - BrcodePayment pdf file
    """
    return rest.get_pdf(resource=_resource, id=id, user=user)


def query(limit=None, after=None, before=None, tags=None, ids=None, status=None, user=None):
    """# Retrieve BrcodePayments
    Receive a generator of BrcodePayment objects previously created in the Stark Bank API
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - after [datetime.date or string, default None] date filter for objects created only after specified date. ex: datetime.date(2020, 3, 10)
    - before [datetime.date or string, default None] date filter for objects created only before specified date. ex: datetime.date(2020, 3, 10)
    - tags [list of strings, default None]: tags to filter retrieved objects. ex: ["tony", "stark"]
    - ids [list of strings, default None]: list of ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - status [string, default None]: filter for status of retrieved objects. ex: "success"
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of BrcodePayment objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        limit=limit,
        after=check_date(after),
        before=check_date(before),
        tags=tags,
        ids=ids,
        status=status,
        user=user,
    )


def update(id, status=None, user=None):
    """# Update BrcodePayment entity
    Update an BrcodePayment by passing id, if it hasn't been paid yet.
    ## Parameters (required):
    - id [string]: BrcodePayment id. ex: '5656565656565656'
    ## Parameters (required):
    - status [string]: You may cancel the payment by passing 'canceled' in the status
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - target BrcodePayment with updated attributes
    """
    return rest.patch_id(resource=_resource, id=id, user=user, status=status)





class Enum:

    @classmethod
    def values(cls):
        return [attribute for attribute in dir(cls) if not attribute.startswith('_') and not callable(getattr(cls, attribute))]

    @classmethod
    def is_valid(cls, item):
        return item in cls.values()



version = "2.4.0"

user = None
language = "en-US"
timeout = 15

from .user.__organization import Organization
from .user.__project import Project

from . import transaction
from .transaction.__transaction import Transaction

from . import balance
from .balance.__balance import Balance

from . import boleto
from .boleto.__boleto import Boleto

from . import transfer
from .transfer.__transfer import Transfer

from . import boletoholmes
from .boletoholmes.__boletoholmes import BoletoHolmes

from . import brcodepayment
from .brcodepayment.__brcodepayment import BrcodePayment

from . import brcodepreview
from .brcodepreview.__brcodepreview import BrcodePreview

from . import boletopayment
from .boletopayment.__boletopayment import BoletoPayment

from . import utilitypayment
from .utilitypayment.__utilitypayment import UtilityPayment

from . import webhook
from .webhook.__webhook import Webhook

from . import workspace
from .workspace.__workspace import Workspace

from . import event
from .event.__event import Event

from . import paymentrequest
from .paymentrequest.__paymentrequest import PaymentRequest

from . import invoice
from .invoice.__invoice import Invoice

from . import dictkey
from .dictkey.__dictkey import DictKey

from . import deposit
from .deposit.__deposit import Deposit

from . import error
from . import key



from ...utils import rest
from ...utils.checks import check_datetime, check_date
from ...utils.api import from_api_json
from ...utils.resource import Resource
from ..__boleto import _resource as _boleto_resource


class Log(Resource):
    """# boleto.Log object
    Every time a Boleto entity is updated, a corresponding boleto.Log
    is generated for the entity. This log is never generated by the
    user, but it can be retrieved to check additional information
    on the Boleto.
    ## Attributes:
    - id [string]: unique id returned when the log is created. ex: "5656565656565656"
    - boleto [Boleto]: Boleto entity to which the log refers to.
    - errors [list of strings]: list of errors linked to this Boleto event
    - type [string]: type of the Boleto event which triggered the log creation. ex: "registered" or "paid"
    - created [datetime.datetime]: creation datetime for the log. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    """

    def __init__(self, id, created, type, errors, boleto):
        Resource.__init__(self, id=id)

        self.created = check_datetime(created)
        self.type = type
        self.errors = errors
        self.boleto = from_api_json(_boleto_resource, boleto)


_resource = {"class": Log, "name": "BoletoLog"}


def get(id, user=None):
    """# Retrieve a specific boleto.Log
    Receive a single boleto.Log object previously created by the Stark Bank API by its id
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - boleto.Log object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def query(limit=None, after=None, before=None, types=None, boleto_ids=None, user=None):
    """# Retrieve boleto.Log's
    Receive a generator of boleto.Log objects previously created in the Stark Bank API
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - after [datetime.date or string, default None] date filter for objects created only after specified date. ex: datetime.date(2020, 3, 10)
    - before [datetime.date or string, default None] date filter for objects created only before specified date. ex: datetime.date(2020, 3, 10)
    - types [list of strings, default None]: filter for log event types. ex: "paid" or "registered"
    - boleto_ids [list of strings, default None]: list of Boleto ids to filter logs. ex: ["5656565656565656", "4545454545454545"]
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - list of boleto.Log objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        limit=limit,
        after=check_date(after),
        before=check_date(before),
        types=types,
        boleto_ids=boleto_ids,
        user=user,
    )



from ..utils import rest
from ..utils.checks import check_date, check_datetime
from ..utils.resource import Resource


class Deposit(Resource):

    """# Deposit object
    Deposits represent passive cash-in received by your account from external transfers
    ## Attributes (return-only):
    - id [string]: unique id associated with a Deposit when it is created. ex: "5656565656565656"
    - name [string]: payer name. ex: "Iron Bank S.A."
    - tax_id [string]: payer tax ID (CPF or CNPJ). ex: "012.345.678-90" or "20.018.183/0001-80"
    - bank_code [string]: payer bank code in Brazil. ex: "20018183" or "341"
    - branch_code [string]: payer bank account branch. ex: "1357-9"
    - account_number [string]: payer bank account number. ex: "876543-2"
    - account_type [string]: payer bank account type. ex: "checking"
    - amount [integer]: Deposit value in cents. ex: 1234 (= R$ 12.34)
    - type [string]: Type of settlement that originated the deposit. ex: "pix" or "ted"
    - status [string]: current Deposit status. ex: "created"
    - tags [list of strings]: list of strings that are tagging the deposit. ex: ["reconciliationId", "txId"]
    - fee [integer]: fee charged by this deposit. ex: 50 (= R$ 0.50)
    - transaction_ids [list of strings]: ledger transaction ids linked to this deposit (if there are more than one, all but first are reversals). ex: ["19827356981273"]
    - created [datetime.datetime]: creation datetime for the Deposit. ex: datetime.datetime(2020, 12, 10, 10, 30, 0, 0)
    - updated [datetime.datetime]: latest update datetime for the Deposit. ex: datetime.datetime(2020, 12, 10, 10, 30, 0, 0)
    """

    def __init__(self, id, name, tax_id, bank_code, branch_code, account_number, account_type, amount, type, status,
                 tags, fee, transaction_ids, created, updated):
        Resource.__init__(self, id=id)

        self.name = name
        self.tax_id = tax_id
        self.bank_code = bank_code
        self.branch_code = branch_code
        self.account_number = account_number
        self.account_type = account_type
        self.amount = amount
        self.type = type
        self.status = status
        self.tags = tags
        self.fee = fee
        self.transaction_ids = transaction_ids
        self.created = check_datetime(created)
        self.updated = check_datetime(updated)


_resource = {"class": Deposit, "name": "Deposit"}


def get(id, user=None):
    """# Retrieve a specific Deposit
    Receive a single Deposit object from the Stark Bank API by its id
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Deposit object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def query(limit=None, after=None, before=None, status=None, sort=None, tags=None, ids=None, user=None):
    """# Retrieve Deposits
    Receive a generator of Deposit objects from the Stark Bank API
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - after [datetime.date or string, default None] date filter for objects created only after specified date. ex: datetime.date(2020, 3, 10)
    - before [datetime.date or string, default None] date filter for objects created only before specified date. ex: datetime.date(2020, 3, 10)
    - status [string, default None]: filter for status of retrieved objects. ex: "paid" or "registered"
    - sort [string, default "-created"]: sort order considered in response. Valid options are "created" or "-created".
    - tags [list of strings, default None]: tags to filter retrieved objects. ex: ["tony", "stark"]
    - ids [list of strings, default None]: list of ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of Deposit objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        limit=limit,
        after=check_date(after),
        before=check_date(before),
        sort=sort,
        status=status,
        tags=tags,
        ids=ids,
        user=user,
    )



from ..utils import rest
from ..utils.checks import check_date, check_datetime
from ..utils.resource import Resource


class Boleto(Resource):
    """# Boleto object
    When you initialize a Boleto, the entity will not be automatically
    sent to the Stark Bank API. The 'create' function sends the objects
    to the Stark Bank API and returns the list of created objects.
    ## Parameters (required):
    - amount [integer]: Boleto value in cents. Minimum = 200 (R$2,00). ex: 1234 (= R$ 12.34)
    - name [string]: payer full name. ex: "Anthony Edward Stark"
    - tax_id [string]: payer tax ID (CPF or CNPJ) with or without formatting. ex: "01234567890" or "20.018.183/0001-80"
    - street_line_1 [string]: payer main address. ex: Av. Paulista, 200
    - street_line_2 [string]: payer address complement. ex: Apto. 123
    - district [string]: payer address district / neighbourhood. ex: Bela Vista
    - city [string]: payer address city. ex: Rio de Janeiro
    - state_code [string]: payer address state. ex: GO
    - zip_code [string]: payer address zip code. ex: 01311-200
    ## Parameters (optional):
    - due [datetime.date or string, default today + 2 days]: Boleto due date in ISO format. ex: 2020-04-30
    - fine [float, default 0.0]: Boleto fine for overdue payment in %. ex: 2.5
    - interest [float, default 0.0]: Boleto monthly interest for overdue payment in %. ex: 5.2
    - overdue_limit [integer, default 59]: limit in days for payment after due date. ex: 7 (max: 59)
    - descriptions [list of dictionaries, default None]: list of dictionaries with "text":string and (optional) "amount":int pairs
    - discounts [list of dictionaries, default None]: list of dictionaries with "percentage":float and "date":datetime.datetime or string pairs
    - tags [list of strings]: list of strings for tagging
    - receiver_name [string]: receiver (Sacador Avalista) full name. ex: "Anthony Edward Stark"
    - receiver_tax_id [string]: receiver (Sacador Avalista) tax ID (CPF or CNPJ) with or without formatting. ex: "01234567890" or "20.018.183/0001-80"
    ## Attributes (return-only):
    - id [string, default None]: unique id returned when Boleto is created. ex: "5656565656565656"
    - fee [integer, default None]: fee charged when Boleto is paid. ex: 200 (= R$ 2.00)
    - line [string, default None]: generated Boleto line for payment. ex: "34191.09008 63571.277308 71444.640008 5 81960000000062"
    - bar_code [string, default None]: generated Boleto bar-code for payment. ex: "34195819600000000621090063571277307144464000"
    - status [string, default None]: current Boleto status. ex: "registered" or "paid"
    - created [datetime.datetime, default None]: creation datetime for the Boleto. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    - our_number [string, default None]: Reference number registered at the settlement bank. ex:"10131474"
    """

    def __init__(self, amount, name, tax_id, street_line_1, street_line_2, district, city, state_code, zip_code,
                 due=None, fine=None, interest=None, overdue_limit=None, tags=None, descriptions=None, discounts=None,
                 receiver_name=None, receiver_tax_id=None, id=None, fee=None, line=None, bar_code=None, status=None,
                 created=None, our_number=None):
        Resource.__init__(self, id=id)

        self.amount = amount
        self.fee = fee
        self.name = name
        self.tax_id = tax_id
        self.street_line_1 = street_line_1
        self.street_line_2 = street_line_2
        self.district = district
        self.city = city
        self.state_code = state_code
        self.zip_code = zip_code
        self.receiver_name = receiver_name
        self.receiver_tax_id = receiver_tax_id
        self.due = check_date(due)
        self.fine = fine
        self.interest = interest
        self.overdue_limit = overdue_limit
        self.tags = tags
        self.descriptions = descriptions
        self.discounts = discounts
        self.line = line
        self.bar_code = bar_code
        self.status = status
        self.created = check_datetime(created)
        self.our_number = our_number


_resource = {"class": Boleto, "name": "Boleto"}


def create(boletos, user=None):
    """# Create Boletos
    Send a list of Boleto objects for creation in the Stark Bank API
    ## Parameters (required):
    - boletos [list of Boleto objects]: list of Boleto objects to be created in the API
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - list of Boleto objects with updated attributes
    """
    return rest.post_multi(resource=_resource, entities=boletos, user=user)


def get(id, user=None):
    """# Retrieve a specific Boleto
    Receive a single Boleto object previously created in the Stark Bank API by its id
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Boleto object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def pdf(id, layout=None, hidden_fields=None, user=None):
    """# Retrieve a specific Boleto pdf file
    Receive a single Boleto pdf file generated in the Stark Bank API by its id.
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - layout [string]: Layout specification. Available options are "default" and "booklet"
    - hidden_fields [list of strings, default None]: List of string fields to be hidden in Boleto pdf. ex: ["customerAddress"]
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Boleto pdf file
    """
    return rest.get_pdf(resource=_resource, id=id, layout=layout, hidden_fields=hidden_fields, user=user)


def query(limit=None, status=None, tags=None, ids=None, after=None, before=None, user=None):
    """# Retrieve Boletos
    Receive a generator of Boleto objects previously created in the Stark Bank API
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - after [datetime.date or string, default None] date filter for objects created only after specified date. ex: datetime.date(2020, 3, 10)
    - before [datetime.date or string, default None] date filter for objects created only before specified date. ex: datetime.date(2020, 3, 10)
    - status [string, default None]: filter for status of retrieved objects. ex: "paid" or "registered"
    - tags [list of strings, default None]: tags to filter retrieved objects. ex: ["tony", "stark"]
    - ids [list of strings, default None]: list of ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of Boleto objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        limit=limit,
        after=check_date(after),
        before=check_date(before),
        status=status,
        tags=tags,
        ids=ids,
        user=user,
    )


def delete(id, user=None):
    """# Delete a Boleto entity
    Delete a Boleto entity previously created in the Stark Bank API
    ## Parameters (required):
    - id [string]: Boleto unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - deleted Boleto object
    """
    return rest.delete_id(resource=_resource, id=id, user=user)



from .__user import User


class Organization(User):
    """# Organization object
    The Organization object is an authentication entity for the SDK that
    represents your entire Organization, being able to access any Workspace
    underneath it and even create new Workspaces. Only a legal representative
    of your organization can register or change the Organization credentials.
    All requests to the Stark Bank API must be authenticated via an SDK user,
    which must have been previously created at the Stark Bank website
    [https://sandbox.web.starkbank.com] or [https://web.starkbank.com]
    before you can use it in this SDK. Organizations may be passed as the user parameter on
    each request or may be defined as the default user at the start (See README).
    If you are accessing a specific Workspace using Organization credentials, you should
    specify the workspace ID when building the Organization object or by request, using
    the Organization.replace(organization, workspace_id) function, which creates a copy of the organization
    object with the altered workspace ID. If you are listing or creating new Workspaces, the
    workspace_id should be None.
    ## Parameters (required):
    - id [string]: unique id required to identify organization. ex: "5656565656565656"
    - private_key [EllipticCurve.Organization()]: PEM string of the private key linked to the organization. ex: "-----BEGIN PUBLIC KEY-----\nMFYwEAYHKoZIzj0CAQYFK4EEAAoDQgAEyTIHK6jYuik6ktM9FIF3yCEYzpLjO5X/\ntqDioGM+R2RyW0QEo+1DG8BrUf4UXHSvCjtQ0yLppygz23z0yPZYfw==\n-----END PUBLIC KEY-----"
    - environment [string]: environment where the organization is being used. ex: "sandbox" or "production"
    - workspace_id [string]: unique id of the accessed Workspace, if any. ex: None or "4848484848484848"
    ## Attributes (return-only):
    - pem [string]: private key in pem format. ex: "-----BEGIN PUBLIC KEY-----\nMFYwEAYHKoZIzj0CAQYFK4EEAAoDQgAEyTIHK6jYuik6ktM9FIF3yCEYzpLjO5X/\ntqDioGM+R2RyW0QEo+1DG8BrUf4UXHSvCjtQ0yLppygz23z0yPZYfw==\n-----END PUBLIC KEY-----"
    """

    def __init__(self, id, environment, private_key, workspace_id=None):
        self.workspace_id = workspace_id

        User.__init__(
            self,
            id=id,
            private_key=private_key,
            environment=environment,
        )

    def access_id(self):
        if self.workspace_id:
            return "organization/{id}/workspace/{workspace}".format(id=self.id, workspace=self.workspace_id)
        return "organization/{id}".format(id=self.id)

    def replace(organization, workspace_id):
        return Organization(
            id=organization.id,
            environment=organization.environment,
            private_key=organization.pem,
            workspace_id=workspace_id,
        )



from ..utils import rest
from ..utils.checks import check_datetime, check_date
from ..utils.resource import Resource


class Transaction(Resource):
    """# Transaction object
    A Transaction is a transfer of funds between workspaces inside Stark Bank.
    Transactions created by the user are only for internal transactions.
    Other operations (such as transfer or charge-payment) will automatically
    create a transaction for the user which can be retrieved for the statement.
    When you initialize a Transaction, the entity will not be automatically
    created in the Stark Bank API. The 'create' function sends the objects
    to the Stark Bank API and returns the list of created objects.
    ## Parameters (required):
        amount [integer]: amount in cents to be transferred. ex: 1234 (= R$ 12.34)
        description [string]: text to be displayed in the receiver and the sender statements (Min. 10 characters). ex: "funds redistribution"
        external_id [string]: unique id, generated by user, to avoid duplicated transactions. ex: "transaction ABC 2020-03-30"
        received_id [string]: unique id of the receiving workspace. ex: "5656565656565656"
    ## Parameters (optional):
        tags [list of strings]: list of strings for reference when searching transactions (may be empty). ex: ["abc", "test"]
    ## Attributes (return-only):
        sender_id [string]: unique id of the sending workspace. ex: "5656565656565656"
        source [string, default None]: locator of the entity that generated the transaction. ex: "charge/1827351876292", "transfer/92873912873/chargeback"
        id [string, default None]: unique id returned when Transaction is created. ex: "7656565656565656"
        fee [integer, default None]: fee charged when transaction is created. ex: 200 (= R$ 2.00)
        balance [integer, default None]: account balance after transaction was processed. ex: 100000000 (= R$ 1,000,000.00)
        created [datetime.datetime, default None]: creation datetime for the transaction. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    """

    def __init__(self, amount, description, external_id, receiver_id, sender_id=None, tags=None, id=None, fee=None, created=None, source=None, balance=None):
        Resource.__init__(self, id=id)

        self.amount = amount
        self.description = description
        self.tags = tags
        self.external_id = external_id
        self.receiver_id = receiver_id
        self.sender_id = sender_id
        self.fee = fee
        self.created = check_datetime(created)
        self.source = source
        self.balance = balance


_resource = {"class": Transaction, "name": "Transaction"}


def create(transactions, user=None):
    """# Create Transactions
    Send a list of Transaction objects for creation in the Stark Bank API
    ## Parameters (required):
    - transactions [list of Transaction objects]: list of Transaction objects to be created in the API
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - list of Transaction objects with updated attributes
    """
    return rest.post_multi(resource=_resource, entities=transactions, user=user)


def get(id, user=None):
    """# Retrieve a specific Transaction
    Receive a single Transaction object previously created in the Stark Bank API by its id
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Transaction object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def query(limit=None, after=None, before=None, tags=None, external_ids=None, ids=None, user=None):
    """# Retrieve Transactions
    Receive a generator of Transaction objects previously created in the Stark Bank API
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - after [datetime.date or string, default None] date filter for objects created only after specified date. ex: datetime.date(2020, 3, 10)
    - before [datetime.date or string, default None] date filter for objects created only before specified date. ex: datetime.date(2020, 3, 10)
    - tags [list of strings, default None]: tags to filter retrieved objects. ex: ["tony", "stark"]
    - external_ids [list of strings, default None]: list of external ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - ids [list of strings, default None]: list of ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of Transaction objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        limit=limit,
        after=check_date(after),
        before=check_date(before),
        tags=tags,
        external_ids=external_ids,
        ids=ids,
        user=user,
    )



from re import compile


pattern = compile(r"(?<!^)(?=[A-Z0-9])")


def camel_to_snake(string):
    return pattern.sub("_", string).lower()


def snake_to_camel(string):
    split = string.split('_')
    return split[0] + "".join(word.title() for word in split[1:])


def camel_to_kebab(string):
    return pattern.sub("-", string).lower()



from starkbank.utils.checks import check_date
from ..utils import rest
from ..utils.checks import check_datetime
from ..utils.resource import Resource


class DictKey(Resource):

    """# DictKey object
    DictKey represents a PIX key registered in Bacen's DICT system.
    
    ## Parameters (optional):
    - id [string]: DictKey object unique id. ex: "tony@starkbank.com", "722.461.430-04", "20.018.183/0001-80", "+5511988887777", "b6295ee1-f054-47d1-9e90-ee57b74f60d9"
    
    ## Attributes (return-only):
    - type [string, default None]: DICT key type. ex: "email", "cpf", "cnpj", "phone" or "evp"
    - name [string, default None]: key owner full name. ex: "Tony Stark"
    - tax_id [string, default None]: key owner tax ID (CNPJ or masked CPF). ex: "***.345.678-**" or "20.018.183/0001-80"
    - owner_type [string, default None]: DICT key owner type. ex "naturalPerson" or "legalPerson"
    - ispb [string, default None]: bank ISPB associated with the DICT key. ex: "20018183"
    - branch_code [string, default None]: bank account branch code associated with the DICT key. ex: "9585"
    - account_number [string, default None]: bank account number associated with the DICT key. ex: "9828282578010513"
    - account_type [string, default None]: bank account type associated with the DICT key. ex: "checking", "saving" e "salary"
    - status [string, default None]: current DICT key status. ex: "created", "registered", "canceled" or "failed"
    - account_created [datetime.datetime, default None]: creation datetime of the bank account associated with the DICT key. ex: datetime.date(2020, 1, 12, 11, 14, 8)
    - owned [datetime.datetime, default None]: datetime since when the current owner holds this DICT key. ex: datetime.date(2020, 11, 16, 8, 12, 11)
    - created [datetime.datetime, default None]: creation datetime for the DICT key. ex: datetime.date(2020, 11, 16, 8, 12, 11)
    """
    
    def __init__(self, id=None, type=None, name=None, tax_id=None, owner_type=None, ispb=None, branch_code=None,
                 account_number=None, account_type=None, status=None, account_created=None, owned=None, created=None):
        Resource.__init__(self, id=id)

        self.id = id
        self.type = type
        self.name = name
        self.tax_id = tax_id
        self.owner_type = owner_type
        self.ispb = ispb
        self.branch_code = branch_code
        self.account_number = account_number
        self.account_type = account_type
        self.status = status
        self.account_created = check_datetime(account_created)
        self.owned = check_datetime(owned)
        self.created = check_datetime(created)


_resource = {"class": DictKey, "name": "DictKey"}


def get(id, user=None):
    """# Retrieve a specific DictKey
    Receive a single DictKey object by its id
    ## Parameters (required):
    - id [string]: DictKey object unique id and PIX key itself. ex: "tony@starkbank.com", "722.461.430-04", "20.018.183/0001-80", "+5511988887777", "b6295ee1-f054-47d1-9e90-ee57b74f60d9"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - DictKey object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def query(limit=None, type=None, after=None, before=None, ids=None, status=None, user=None):
    """# Retrieve DictKeys
    Receive a generator of DictKey objects associated with your Stark Bank Workspace
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - type [string, default None]: DictKey type. ex: "cpf", "cnpj", "phone", "email" or "evp"
    - after [datetime.date or string, default None] date filter for objects created only after specified date. ex: datetime.date(2020, 3, 10)
    - before [datetime.date or string, default None] date filter for objects created only before specified date. ex: datetime.date(2020, 3, 10)
    - ids [list of strings, default None]: list of ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - status [string, default None]: filter for status of retrieved objects. ex: "success"
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of DictKey objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        limit=limit,
        type=type,
        after=check_date(after),
        before=check_date(before),
        ids=ids,
        status=status,
        user=user,
    )



from ..utils import rest
from ..utils.checks import check_date, check_datetime, check_timedelta
from ..utils.resource import Resource


class Invoice(Resource):
    """# Invoice object
    When you initialize an Invoice, the entity will not be automatically
    sent to the Stark Bank API. The 'create' function sends the objects
    to the Stark Bank API and returns the list of created objects.
    ## Parameters (required):
    - amount [integer]: Invoice value in cents. Minimum = 0 (any value will be accepted). ex: 1234 (= R$ 12.34)
    - tax_id [string]: payer tax ID (CPF or CNPJ) with or without formatting. ex: "01234567890" or "20.018.183/0001-80"
    - name [string]: payer name. ex: "Iron Bank S.A."

    ## Parameters (optional):
    - due [datetime.datetime or string, default today + 2 days]: Invoice due date in UTC ISO format. ex: "2020-10-28T17:59:26.249976+00:00"
    - expiration [integer or datetime.timedelta, default 5097600 (59 days)]: time interval in seconds between due date and expiration date. ex 123456789
    - fine [float, default 2.0]: Invoice fine for overdue payment in %. ex: 2.5
    - interest [float, default 1.0]: Invoice monthly interest for overdue payment in %. ex: 5.2
    - discounts [list of dictionaries, default None]: list of dictionaries with "percentage":float and "due":datetime.datetime or string pairs
    - tags [list of strings, default None]: list of strings for tagging
    - descriptions [list of dictionaries, default None]: list of dictionaries with "key":string and (optional) "value":string pairs

    ## Attributes (return-only):
    - pdf [string, default None]: public Invoice PDF URL. ex: "https://invoice.starkbank.com/pdf/d454fa4e524441c1b0c1a729457ed9d8"
    - nominal_amount [integer, default None]: Invoice emission value in cents (will change if invoice is updated, but not if it's paid). ex: 400000
    - fine_amount [integer, default None]: Invoice fine value calculated over nominal_amount. ex: 20000
    - interest_amount [integer, default None]: Invoice interest value calculated over nominal_amount. ex: 10000
    - discount_amount [integer, default None]: Invoice discount value calculated over nominal_amount. ex: 3000
    - id [string, default None]: unique id returned when Invoice is created. ex: "5656565656565656"
    - brcode [string, default None]: BR Code for the Invoice payment. ex: "00020101021226800014br.gov.bcb.pix2558invoice.starkbank.com/f5333103-3279-4db2-8389-5efe335ba93d5204000053039865802BR5913Arya Stark6009Sao Paulo6220051656565656565656566304A9A0"
    - status [string, default None]: current Invoice status. ex: "registered" or "paid"
    - fee [integer, default None]: fee charged by this Invoice. ex: 200 (= R$ 2.00)
    - created [datetime.datetime, default None]: creation datetime for the Invoice. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    - updated [datetime.datetime, default None]: latest update datetime for the Invoice. ex: datetime.datetime(2020, 3, 10, 10, 30, 0, 0)
    """

    def __init__(self, amount, tax_id, name, due=None, expiration=None, fine=None, interest=None, discounts=None,
                 tags=None, descriptions=None, pdf=None, nominal_amount=None, fine_amount=None, interest_amount=None,
                 discount_amount=None, id=None, brcode=None, status=None, fee=None, created=None, updated=None):
        Resource.__init__(self, id=id)

        self.amount = amount
        self.nominal_amount = nominal_amount
        self.fine_amount = fine_amount
        self.interest_amount = interest_amount
        self.discount_amount = discount_amount
        self.due = check_datetime(due)
        self.tax_id = tax_id
        self.name = name
        self.expiration = check_timedelta(expiration)
        self.fine = fine
        self.interest = interest
        self.discounts = discounts
        self.tags = tags
        self.pdf = pdf
        self.descriptions = descriptions
        self.brcode = brcode
        self.status = status
        self.fee = fee
        self.created = check_datetime(created)
        self.updated = check_datetime(updated)


_resource = {"class": Invoice, "name": "Invoice"}


def create(invoices, user=None):
    """# Create Invoices
    Send a list of Invoice objects for creation in the Stark Bank API
    ## Parameters (required):
    - invoices [list of Invoice objects]: list of Invoice objects to be created in the API
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - list of Invoice objects with updated attributes
    """
    return rest.post_multi(resource=_resource, entities=invoices, user=user)


def get(id, user=None):
    """# Retrieve a specific Invoice
    Receive a single Invoice object previously created in the Stark Bank API by its id
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Invoice object with updated attributes
    """
    return rest.get_id(resource=_resource, id=id, user=user)


def query(limit=None, status=None, tags=None, ids=None, after=None, before=None, user=None):
    """# Retrieve Invoices
    Receive a generator of Invoice objects previously created in the Stark Bank API
    ## Parameters (optional):
    - limit [integer, default None]: maximum number of objects to be retrieved. Unlimited if None. ex: 35
    - after [datetime.date or string, default None] date filter for objects created only after specified date. ex: datetime.date(2020, 3, 10)
    - before [datetime.date or string, default None] date filter for objects created only before specified date. ex: datetime.date(2020, 3, 10)
    - status [string, default None]: filter for status of retrieved objects. ex: "paid" or "registered"
    - tags [list of strings, default None]: tags to filter retrieved objects. ex: ["tony", "stark"]
    - ids [list of strings, default None]: list of ids to filter retrieved objects. ex: ["5656565656565656", "4545454545454545"]
    - user [Project object, default None]: Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - generator of Invoice objects with updated attributes
    """
    return rest.get_list(
        resource=_resource,
        limit=limit,
        after=check_date(after),
        before=check_date(before),
        status=status,
        tags=tags,
        ids=ids,
        user=user,
    )


def update(id, status=None, amount=None, due=None, expiration=None, user=None):
    """# Update Invoice entity
    Update an Invoice by passing id, if it hasn't been paid yet.
    ## Parameters (required):
    - id [string]: Invoice id. ex: '5656565656565656'
    ## Parameters (optional):
    - status [string]: You may cancel the invoice by passing 'canceled' in the status
    - amount [string]: Nominal amount charged by the invoice. ex: 100 (R$1.00)
    - due [datetime.datetime or string, default now + 2 days]: Invoice due date in UTC ISO format. ex: "2020-10-28T17:59:26.249976+00:00"
    - expiration [integer or datetime.timedelta, default None]: time interval in seconds between the due date and the expiration date. ex 123456789
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - target Invoice with updated attributes
    """
    payload = {
        "status": status,
        "amount": amount,
        "due": check_datetime(due),
        "expiration": check_timedelta(expiration),
    }
    return rest.patch_id(resource=_resource, id=id, user=user, **payload)


def qrcode(id, size=7, user=None):
    """# Retrieve a specific Invoice QR Code png
    Receive a single Invoice QR Code in png format generated in the Stark Bank API by the invoice ID.
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - size [integer, default 7]: number of pixels in each "box" of the QR code. Minimum = 1, maximum = 50. ex: 12
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Invoice png blob
    """
    return rest.get_qrcode(resource=_resource, id=id, size=size, user=user)


def pdf(id, user=None):
    """# Retrieve a specific Invoice pdf file
    Receive a single Invoice pdf receipt file generated in the Stark Bank API by its id.
    ## Parameters (required):
    - id [string]: object unique id. ex: "5656565656565656"
    ## Parameters (optional):
    - user [Organization/Project object, default None]: Organization or Project object. Not necessary if starkbank.user was set before function call
    ## Return:
    - Invoice pdf file
    """
    return rest.get_pdf(resource=_resource, id=id, user=user)
